(*  
 *  This file is a component of PropPlan a model-based planner
 *  Copyright (C) 2000-2007  Michael Paul Fourman

 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)
(*****************************************************************************************)
(**** $Id: PDDL.ML,v 1.1 2007/12/30 15:31:25 michaelfourman Exp $ ****)
(*****************************************************************************************)

infix 5 -- ==> ;
infix 3 >>;
infix 0 ||;

infix 4 &&;

functor PDDL (structure Parse : PARSE where type instream = TextIO.StreamIO.instream
	      structure Prop : PROPOSITIONSIG where type Name = string
		  )(*:PDDLSIG*) =
    struct
	open Prop;
	open Parse;
    
	type AtomicFormulaSkeleton = Predicate * Term TypedList

	datatype Effect = EffAtom of Formula                            (* includes simple quantified effects *)
	  | EffWhen of Formula * Formula
	  | EffForall of Term TypedList * (Formula  * Formula)          (* quantified when *)
	  | EffAnd of Effect list

	fun predicatesInEffect (EffAtom p)           = predicatesInFormula p
	  | predicatesInEffect (EffWhen(_, p))       = predicatesInFormula p
	  | predicatesInEffect (EffForall(_,(_, p))) = predicatesInFormula p
	  | predicatesInEffect (EffAnd ps)           = uniquePredicates (map predicatesInEffect ps)

	datatype ActionDefBody =
	    Precondition of Formula
	  | Effect of Effect
	  | CondEffect of Formula * Formula

	fun getEffect [] = raise SynError ("Action with no effect",[])
	  | getEffect (Effect e :: t) =  e 
	  | getEffect (_ :: t) = getEffect t

	fun getPrecondition [] = T
	  | getPrecondition (Precondition e :: t) = e && getPrecondition t
	  | getPrecondition (_ :: t) = getPrecondition t

	datatype Action = Action of {name: Name,
				     parameters: Term TypedList,
				     precondition: Formula, 
				     effect: Effect }
	    
	fun actEffect (Action{effect,...}) = effect

	fun mkAction ((name, parameters), defs) = 
	    Action {name = name,
		    parameters = parameters, 
		    precondition = getPrecondition defs, 
		    effect = getEffect defs}

	datatype DomainDef =
	    Types of Name TypedList
	  | Constants of Term TypedList
	  | Predicates of AtomicFormulaSkeleton list
	  | Require of Name list
          | Axiom of Formula
	  | Actions of Action list


	fun getTypes  [] = TypedList []
	  | getTypes  (Types tl :: t) = tl @@ getTypes t
	  | getTypes (_ :: t) = getTypes t

	fun getConstants [] = TypedList []
	  | getConstants (Constants tl :: t) = tl @@ getConstants t
	  | getConstants (_ :: t) = getConstants t

	fun getPredicates [] = []
	  | getPredicates (Predicates tl :: t) = tl @ getPredicates t
	  | getPredicates (_ :: t) = getPredicates t

 
	fun getAxiom [] = []
	  | getAxiom (Axiom tl :: t) = tl :: getAxiom t
	  | getAxiom (_ :: t) = getAxiom t

	fun getRequire [] = []
	  | getRequire (Require tl :: t) = tl @ getRequire t
	  | getRequire (_ :: t) = getRequire t

	fun getActions [] = []
	  | getActions (Actions tl :: t) = tl @ getActions t
	  | getActions (_ :: t) = getActions t


	datatype Domain = Domain of {name: Name,
				     types: Name TypedList, 
				     constants: Term TypedList, 
				     predicates: AtomicFormulaSkeleton list, 
				     require: Name list,
				     axiom: Formula,
				     actions: Action list}


	fun mkDomain (name, defs) = Domain{name = name,
					   types= getTypes defs, 
					   constants = getConstants defs, 
					   predicates = getPredicates defs,
					   axiom = &&&(getAxiom defs),
					   require = getRequire defs,
					   actions = getActions defs}


	fun K x y = x

	fun seq p toks = (p -- repeat p >> op ::) toks

	val name = id 

	fun literal t toks =
	    ( atomic_formula t                                
	  ||  $"(" :- $"not" :- atomic_formula t -: $")"     >> Prop.~
	      ) toks
	and atomic_formula t toks = 
	    ( $"(" :- predicate -- repeat t -: $")"          >> (Prop.A o Prop.mkLit)
	     ) toks
	and term toks =
	    ( name                                           >> ^^
	  ||  variable                                          
	      ) toks
	and predicate toks =
	    ( name                                           >> %
	  ||  $"="                                           >> K (op ==)
	      ) toks
	and variable toks =
	    ( $"?" :- name                                   >> ? 
	     ) toks 

	(*--- ADL stuff put here by rwd 3/3/2000 -----------*)

	fun pddltype toks = 
	    ( name                                            >> Type
	  ||  $"(" :- $"either" :- seq pddltype -: $")"       >> Either
	    ||  $"object"                                       >> K Object
		) toks

	(*PDDL Types *)
	(*NB the PDDL subset definition (DRAFT 1 Feb14) appears flawed here *)
	fun typed x toks =
	    ( (seq x -: $"-" -- pddltype) 
||   seq x                                          >> (fn x => (x,Object))
	     ) toks 

	fun typed_list x toks = 
	    (
	     repeat (typed x)                                        >> TypedList
	     ) toks 

	(********************************  Figure 3 ************************************************************)
	fun formula toks = 
	    (    literal term
	  || $"(" :- $"not" :- formula -: $")"                      >> Prop.~ 
	   || $"(" :- $"and" :- repeat formula -: $")"               >> Prop.&&& (* don't insist on *)
	    || $"(" :- $"or"  :- repeat formula -: $")"               >> Prop.||| (* a sequence here *)
|| $"(" :- $"imply" :- formula -- formula  -: $")"        >> Prop.==>
|| $"(" :- $"iff" :- formula -- formula  -: $")"          >> Prop.<=>
     || $"(" :- $"exists" :- $"(" :- typed_list variable -: $")" -- formula  -: $")" >> Prop.EXS
      || $"(" :- $"forall" :- $"(" :- typed_list variable -: $")" -- formula  -: $")" >> Prop.ALL
       ||  $"(" :- $":" :- $"goal" :- formula -: $")"            >> GOAL
	 ||  $"(" :- $":" :- $"init" :- formula -: $")"            >> INIT
	   || $"(" :- $"<" :- ## -- $"(" :- typed_list variable -: $")" -- formula  -: $")" >> Prop.LT
	    || $"(" :- $"=" :- ## -- $"(" :- typed_list variable -: $")" -- formula  -: $")" >> Prop.EQ
             || $"(" :- $">" :- ## -- $"(" :- typed_list variable -: $")" -- formula  -: $")" >> Prop.GT
	      || $"(" :- $"=<" :- ## -- $"(" :- typed_list variable -: $")" -- formula  -: $")" >> Prop.LE
	       || $"(" :- $">=" :- ## -- $"(" :- typed_list variable -: $")" -- formula  -: $")" >> Prop.GE
		  ) toks

	fun when_eff toks = 
	    (  $"(" :- $"when"  :- formula -- formula  -: $")"           
	     ) toks

	fun eff_formula toks = 
	    (  when_eff                                                                   >> EffWhen
	  || $"(" :- $"forall":- $"(" :- typed_list variable -: $")" -- when_eff  -: $")" >> EffForall 
	   || $"(" :- $"and"   :- seq eff_formula  -: $")"                                 >> EffAnd
	    || formula                                                                      >> EffAtom
	       ) toks

	fun actiondef toks =  
	    ( $":" :- $"effect"  :- eff_formula                         >> Effect
	  ||  $":" :- $"precondition"  :- formula                       >> Precondition
	      ) toks

	fun action toks = 
	    (( $"(" :- $":" :- $"action"  :- name 
	      -- ($":" :- $"parameters" :- $"(" :- typed_list variable -: $")"
	    || empty >> TypedList)                 (* not in spec -- but needed for examples *)
	      -- (actiondef -- repeat actiondef >> op ::)
	      -: $")" 
	      ) >> mkAction) toks


	fun atomic_formula_skeleton toks =
	    ( $"(" :- predicate -- typed_list variable -: $")"  
	     ) toks

	fun require_key toks =
	    (  $":" :- $"strips"
	  || $":" :- $"adl"
	   || $":" :- $"typing"
	      ) toks

	fun def toks = 
	    ( $"(" :- $":" :- $"requirements" :- seq require_key -: $")" >> Require
	  ||  $"(" :- $":" :- $"types" :- typed_list name -: $")"        >> Types
	  ||  $"(" :- $":" :- $"constants" :- typed_list term -: $")"    >> Constants
	  ||  $"(" :- $":" :- $"axiom" :- formula -: $")"                >> Axiom
	  ||  $"(" :- $":" :- $"predicates" :- repeat atomic_formula_skeleton -: $")"     >> Predicates
	  ||  seq action                                                 >> Actions
		      ) toks

	fun domain toks =
	    ( $"(" :- $"define" :- $"(" :- $"domain" :- name -: $")" 
	     -- repeat def 
	     -: $")" >> mkDomain
	     ) toks

	datatype ProblemDef = 
	    Init of Formula list
	  | Objects of Term TypedList
	  | Goal of Formula list
	  | Junk 

	fun getInit [] = []
	  | getInit (Init tl :: t) = tl @ getInit t
	  | getInit (_ :: t) = getInit t

	fun getObjects [] = TypedList []
	  | getObjects (Objects tl :: t) = tl @@ getObjects t
	  | getObjects (_ :: t) = getObjects t

	fun getGoals [] = []
	  | getGoals (Goal tl :: t) = tl @ getGoals t
	  | getGoals (_ :: t) = getGoals t

	fun problemdef toks =
	    ( $"(" :- $":" :- $"init"    :- repeat formula  -: $")"  >> Init
	  ||  $"(" :- $":" :- $"objects" :- typed_list term -: $")"  >> Objects 
	  ||  $"(" :- $":" :- $"goal"    :- repeat formula  -: $")"  >> Goal
		) toks

	datatype Problem = Problem of 
	    {name: Name, 
	     domain: Name, 
	     objects: Term TypedList, 
	     init: Formula, 
	     goal: Formula}

	datatype Batch = Batch of 
	    {name : Name, 
	     types: Name TypedList, 
	     actions: Action list,
	     require: Name list, 
	     predicates: AtomicFormulaSkeleton list,
	     objects: Term TypedList, 
	     init: Formula,
	     axiom: Formula,
	     goals: (Name * Formula) list}

	fun selectSimilar 
	    (differentProblems,similar)
	    (thisProblem as 
	     Problem {domain, 
		      objects, 
		      init,
		      ...}) 
	    ((otherProblem as
	      Problem {name = otherName, 
		       domain = otherDomain, 
		       objects = otherObjects,
		       init = otherInit, 
		       goal = otherGoal,
                       ...}
	      ) 
	    :: moreProblems)
	    = if( domain = otherDomain 
		 andalso objects = otherObjects 
		 andalso init = otherInit)
		  then
		      selectSimilar (differentProblems, (otherName, otherGoal) :: similar) thisProblem moreProblems
	      else 
		  selectSimilar(otherProblem :: differentProblems,similar) thisProblem moreProblems
	  | selectSimilar (diff,sim) _ [] = (rev diff,sim)

	fun mkBatch  
	    (thisDomain as
	     Domain {name = domainName,
		     types, 
		     constants: Term TypedList, 
		     predicates: AtomicFormulaSkeleton list, 
		     require: Name list, 
		     actions: Action list,
		     axiom: Formula,
		     ...},
	     (firstProblem as
	      Problem {name = problemName, 
		       domain = problemDomain, 
		       objects: Term TypedList, 
		       init: Formula, 
		       goal: Formula,
		       ...})
	     :: moreProblems)
	    = if domainName <> problemDomain then
	    (TextIO.output(TextIO.stdOut, "Problem " ^ problemName 
		    ^ " domainName <> problemDomain !");[])
	      else
		  let val (differentProblems, similar) = 
		      selectSimilar ([],[(problemName, goal)]) firstProblem moreProblems
		  in
		      Batch{name       = domainName, 
			    types      = types, 
			    actions    = actions,
			    axiom      = axiom,
			    require    = require, 
			    predicates = predicates,
			    objects    = objects @@ constants,
			    init       = init,
			    goals      = similar
			    }
		      :: mkBatch (thisDomain, differentProblems)
		  end
	  | mkBatch (_, []) = []

	fun mkProblem ((name, domain), defs) =
	    Problem{name = name,
		    domain = domain,
		    objects = getObjects defs,
		    init = &&&(getInit defs),
		    goal = &&&(getGoals defs)}

	fun problem toks =
	    ( $"(" :- 
	     $"define" :- $"(" :- $"problem" :- name -: $")"    
	     --  $"(" :- $":" :- $"domain" :- name -: $")"
	     -- repeat problemdef 
	     -: $")"  >> mkProblem
	     ) toks  

	fun batch toks = 
	    (
	     ( domain -- seq problem ) >> mkBatch
	     ) toks

	fun readBatch f  = Parse.reader batch f
	fun readDomain f  = Parse.reader domain f
	fun readProblem f = Parse.reader problem f

    end
