(*  
 *  This file is a component of PropPlan - a model-based planner
 *  Copyright (C) 2000-2007  Michael Paul Fourman
 *  homepages.inf.ed.ac.uk/mfourman/tools/propplan
 *  sourceforge.net/projects/propplan
 *  michael.fourman (AT) gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of version 3 of the GNU Affero General Public License
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received the file COPYING, a copy of the GNU Affero 
 *  Public License, along with this program.  
 *  If not, see <http://www.gnu.org/licenses/>.
 
 *  $Id: COMPILER.ML,v 1.2 2007/12/31 19:44:13 michaelfourman Exp $
 *)

infix 3 @@;
infix 2 ==>
infix 3 OR 
infix 4 AND 
infix 5 == <>; 
infix 2 IMPLIES IFF;
infix mem \;
nonfix --;

functor COMPILER(structure Parser : PDDLSIG where type Name = string
		 and Planner   : PLANNERSIG where type Bdd.Name = string
		 and Vars      : DICTIONARY 
                 and Basic     : BASIC
		 sharing type Planner.Bdd.Lit = Parser.Lit
			 sharing type Vars.Key = Planner.Bdd.Var
			 sharing type Parser.Formula = Planner.Bdd.Formula
			 sharing type Planner.Bdd = Planner.Bdd.Bdd
			 sharing type Planner.Vars = Planner.Bdd.Vars
			sharing type Planner.Var = Planner.Bdd.Var
			sharing type Planner.Bdd.TypedList = Parser.TypedList
			sharing type Planner.Bdd.Term = Parser.Term
			    ) =
    struct
	local
	    open Planner.Bdd
	     Parser
	     List Basic
	in
	    (* Utility functions *)
            fun message s = ()  (*TextIO.print s;*)
	    (* the business *)
	    fun mkProblem  
		(Batch{
		       name : Name, 
		       types: Name TypedList, 
		       actions: Action list,
		       axiom: Formula,
                       require: Name list, 
		       predicates: AtomicFormulaSkeleton list,
		       objects: Term TypedList, 
		       init: Formula,
		       goals: (Name * Formula) list
		       }) =
 		let val _ = bdd_restart ()
		    val FF = FALSE() and TT = TRUE()
		    val changedPredicates = uniquePredicates (map (predicatesInEffect o actEffect) actions)
	    
		    val bddInit = eval (K FF) (K FF) types objects (V o lit2var) init
		    (* temporary to compute timeless predicates *)

		    val initPredicates = map var2lit (SUPPORT bddInit)
		    val _ = message ("Init " ^ Int.toString (length initPredicates) ^ " predicates\n");

                    fun isTrueInit instantiated =
			instantiated mem initPredicates andalso
			let val theAtom = V(lit2var instantiated) (* create or reuse  bdd *)
			in
			    ++bddInit IMPLIES theAtom
                	end
                    fun isFalseInit instantiated = (* NB actVars are false initially *)
                        not( instantiated mem initPredicates) orelse
			let val theAtom =  V(lit2var instantiated) (* create or reuse  bdd *)
			in
                            ++bddInit IMPLIES NOT theAtom
                	end
		    fun isInit bddf = if ++bddInit IMPLIES bddf then TT else FF

		    fun testEq [x,y] = if x = y then TT else FF

		    (* translate atoms, using FF or TT for unchanged predicates set by init *)
		    fun translate ss atom =
			let val instantiated as !!(p,xs) =  ss atom
			in  if p = == then
			    testEq xs
			    else
				if p mem changedPredicates then
				    V(lit2var instantiated)
				else
				    if isTrueInit instantiated then TT
				    else
					if isFalseInit instantiated then FF
					else
					    V(lit2var instantiated)
			end

		    val init = eval (K FF) (K FF) types objects (translate I) init

		    val goals = map (fn (s,g) => 
				     (s,eval (K FF) isInit types objects (translate I) g)) 
			goals
		    fun isGoal bddf = 
			let fun isGf [] = false
			      | isGf (g :: gs) = 
				(g IMPLIES ++bddf) orelse isGf gs
			    val r = isGf (map (++ o #2) goals)
			in
			    (--bddf; if r then TT else FF)
			end 

		    val constraint = eval isGoal isInit types objects (translate I) axiom 
		    exception Redundant
		    fun emap f [] = []
		      | emap f (h :: t) = (f h :: emap f t) 
			handle Redundant => emap f t

		    fun action {name: Name, 
				parameters: Term TypedList,
				precondition: Formula,
				effect  : Effect} s =
			let val mkBdd =  eval isGoal isInit types objects (translate (tr s))
			    val effToken = ref ""
			    val id = (tr s (!!(% name, forgetTypes parameters))) 

			    fun mkToken() = lit2var (effToken := "a" ^ !effToken;
						     !!(==,[^^ (!effToken)]))
			    (* a hack -- something that can't be mistaken for a real predicate *)
				
			    fun dispose effects = map (fn (a,p,q) => (--(V a);--p;--q)) effects
			    fun mkCondEff [] ceff = ceff
			      | mkCondEff ((a,p,q) :: t) (pre, chs, eff) =
				if ++pre INTERSECTS ++p andalso ++eff INTERSECTS ++q then
				    mkCondEff t ((++(V a) IFF p) AND pre,
						 (a, Vars(SUPPORT q)) :: chs,
						 (++(V a) IMP q) AND eff) 
				else
				    (--(V a);--p;--q;mkCondEff t (pre,chs,eff))
			    fun expandEffect precond (EffAtom atom) =
				let val theEffect = mkBdd atom
				in  if not(++constraint INTERSECTS ++theEffect) orelse precond = FF then
				    (--precond; --theEffect;raise Redundant)
				    else
					expandEffects [] 
					(precond, REDUCE theEffect (++constraint), [])
				end
			      | expandEffect precond (EffWhen (p,q)) =
				let val pbdd      = mkBdd p
				    and qbdd      = REDUCE (mkBdd q) (++constraint)
				in
				    if not((++precond AND ++pbdd) INTERSECTS ++constraint)
					orelse qbdd = FF then
					(--pbdd;--qbdd;--precond;raise Redundant)
				    else
					expandEffects [] (precond, TT, [(mkToken(), pbdd, qbdd)])
				end
			      | expandEffect precond (EffAnd effs) =
				if not(++precond INTERSECTS ++constraint) then
				    (--precond;raise Redundant)
				else
				    expandEffects effs (precond, TT, [])

			      | expandEffect precond (eff as EffForall _) = 
				if not(++precond INTERSECTS ++constraint) then
				    (--precond; raise Redundant)
				else
				    expandEffects [eff] (precond, TT, [])

			    and expandEffects [] (cond, eff, ceffs) =
				if eff = FF orelse cond = FF then
				    (--cond;--eff;dispose ceffs;raise Redundant)
				else	
				    let val token      = lit2var id	
					val name       = lit2string id
					val changes    = Vars(Planner.Bdd.SUPPORT eff)
					val (pre, chs, eff') = mkCondEff ceffs (cond, [], eff)
					val conditions = map #1 chs
				    in    
					[Planner.Action
					 {name        = name,
					  token       = token,
					  precond     = REDUCE pre (++constraint AND NOT (++eff')),
					  effect      = eff',
					  conditions  = Vars conditions,
					  condChanges = map (fn (a,ch) => (V a, ch)) chs,
					  changes     = changes}]
				    end
			      | expandEffects (EffAtom atom :: es) (cond, eff, ceffs) =
				let val e = mkBdd atom AND eff
				in 
				    if ++e IMPLIES NOT (++constraint) orelse cond = FF then
					(--cond;--e;dispose ceffs;raise Redundant)
				    else
					expandEffects  es (cond, REDUCE e (++constraint), ceffs)
				end
			      | expandEffects (EffWhen (p,q) :: es) (cond, eff, ceffs) =
				let val pbdd = mkBdd p
				    val qbdd = mkBdd q AND ++eff
				in 
				    if ++pbdd AND ++cond = FF orelse ++qbdd IMPLIES NOT (++constraint) then
					(--pbdd;--qbdd;expandEffects  es (cond, eff, ceffs))
				    else
					expandEffects es (cond, eff, (mkToken(), pbdd, REDUCE qbdd (++constraint AND ++eff)) :: ceffs)
				end
			      | expandEffects (EffForall (vars, (p, q)) :: es) (cond, eff, ceffs) =
				let val subs = substitutions types objects vars
				    fun inst t = 
					let val pbdd = eval isGoal isInit types objects (translate (tr (s o t))) p 
					    val qbdd = eval isGoal isInit types objects (translate (tr (s o t))) q
					in if ++pbdd AND ++cond = FF orelse qbdd = FF then
					    (--pbdd;--qbdd; raise Redundant)
					   else
					       (mkToken(), pbdd, REDUCE qbdd (++constraint AND ++eff))
					end
				in
				    expandEffects es (cond, eff, (emap inst subs) @ ceffs)
				end
			    val precond = mkBdd precondition    
			in 
			    if ++precond INTERSECTS ++constraint then
				expandEffect precond effect
			    else
				raise Redundant
			end

		    fun expansions (Parser.Action a) = 
			(REORDER();List.concat (emap (action a) 
					    (substitutions types  objects  (#parameters a))))

		    fun nontriv (Planner.Action {effect,precond,name,...}) =
			if 
			    effect  <> FF
			    andalso 
			    precond <> FF
			    then
				true
			else
			    ( TextIO.print ("\nPruning redundant action " ^ name) ;
			     --precond; --effect; false)
		    val acts = (List.concat (map expansions actions))
		    fun getConditions (Planner.Action{condChanges,...}) = map (++ o #1) condChanges
		    val acts =	(REORDER();filter nontriv acts)
		    val actTokens = map Planner.actionToken acts
		    val ceffTokens = SUPPORT (ANDL(List.concat(map getConditions acts)))
		    val actVars = ceffTokens @ actTokens 
		    fun rmRedundant [] = (--bddInit;[])
		      | rmRedundant (v :: vs) =
			if translate I (var2lit v) = V v then
			    v :: rmRedundant vs
			else
			    (rmVar v; rmRedundant vs)
		    val stateVars = rmRedundant(SUPPORT (noVars()) \ actVars)
		    val restfalse = EXISTS (Vars actVars) (EXISTS (Vars(SUPPORT init)) (noVars()))
		(* look for invariants *)
		in  
		    Planner.CompiledProblem
		    {
		     acts = acts,
		     stateVars = Vars stateVars,
		     goals = goals ,
		     init = init AND restfalse AND ++constraint, (* init should satisfy constraint?? *)
		     constraint = constraint,
		     name = name
		     }
		end
	end
    end;
