(*
 *  This file is a component of PropPlan - a model-based planner
 *  Copyright (C) 2000-2007  Michael Paul Fourman
 *  homepages.inf.ed.ac.uk/mfourman/tools/propplan
 *  sourceforge.net/projects/propplan
 *  michael.fourman (AT) gmail.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of version 3 of the GNU Affero General Public License
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received the file COPYING, a copy of the GNU Affero 
 *  Public License, along with this program.  
 *  If not, see <http://www.gnu.org/licenses/>.
 
 *  $Id: ml_bind.ML,v 1.2 2007/12/31 19:44:13 michaelfourman Exp $
 *) 

(* Code to build the top-level system *)

structure PropPlan =
    struct
	local
	    structure Parser   = Parser
	    structure Planner  = Planner
	    structure Compiler = COMPILER( structure Parser  = Parser
					   structure Planner = Planner
					   structure Basic   = BASIC()
					   structure Vars    = DICTIONARY(type Key = Planner.Bdd.Var ))
	in
            val time = ref 600 (*seconds*);
	    type pddlDomain = Parser.Domain;
	    type pddlProblem = Parser.Problem;
	    type CompiledProblem = Planner.CompiledProblem;
	    type Batch = Parser.Batch
	    type Action = Planner.Action
	    val bdd_start = Planner.Bdd.bdd_start
                
	    val bdd_restart = Planner.Bdd.bdd_restart
	    exception Error = Planner.Bdd.Error
	    exception SynError = Parser.SynError
	    val readProblem = Parser.readProblem
	    val readDomain = Parser.readDomain
	    (* TODO make exception "Error of string" visible *)
	    val mkBatch = Parser.mkBatch
	    val mkPlan = Planner.mkPlan
	    val mkProblem = Compiler.mkProblem
	    fun length [] = 0 | length (_ :: t) = 1 + length t;


	    fun process timeOut batch =  
		let val start  =  Timer.startCPUTimer();
		    fun CPUtime() = let val {usr,...} = Timer.checkCPUTimer start in usr end
		    fun timeout () = CPUtime() > Time.fromSeconds timeOut
			andalso (TextIO.print ("\n<-START->\n <b>Timeout</b></h3> " 
                                ^ "<dl><dt>CPU time</dt><dd>"
                                ^ Time.toString (CPUtime()) 
                                ^ "\n</dd></dl>\n<-END->")
				 ;true)
                    fun htmlList (x :: y :: z :: (xs as _ :: _)) = x ^ " " ^ y ^ " " ^ z ^ "<br />" ^ htmlList xs
                      | htmlList (x :: (xs as _ :: _)) =  x ^ " " ^ htmlList xs
                      | htmlList [x] = x
                      | htmlList [] = ""
		    fun sumTimes[] = 0 | sumTimes ((_,t,_):: rs) = t + sumTimes rs
		    val results = Planner.mkPlan timeout (Compiler.mkProblem batch)
		    val timeTaken = CPUtime()
		    val totalTime = sumTimes results
		    fun timeAllotted timeStamp = timeTaken
		    (*real (timeTaken * (timeStamp - start)) / real (totalTime * 10)handle Quot => 0.09*)
		    fun printResult (problemName,timeStamp,actions) = 
			((*<?xml version="1.0" encoding="ISO-8859-1" standalone="yes"?>?>*)
			 TextIO.print("\n<-START->\n"
                                      ^ " <b>" ^ problemName ^ "</b></h3>  "
                                      ^ "<p><b>CPU time</b> "
                                      ^ PolyML.makestring (timeAllotted timeStamp) ^ " <b>Plan length</b> "
                                      ^ PolyML.makestring(length actions)
                          ^ "</p><p> "
                          ^ htmlList (map Planner.actionName actions)
                          ^"</p>\n<-END->");
			 TextIO.flushOut  TextIO.stdOut)
		in 
		    map printResult results
		end

	    fun propplan domainName problemNames =
		let val domainFile = TextIO.openIn domainName
		    val problemFiles = map TextIO.openIn problemNames
		    val domain = readDomain (TextIO.getInstream domainFile)
		    val problems = map (readProblem o TextIO.getInstream) problemFiles
		    val batch = mkBatch (domain, problems)
		in 
		    (TextIO.closeIn domainFile;
                     map TextIO.closeIn problemFiles;
                     map (process (!time)) batch)
		end

            (* takes domain and problems on st_in outputs results on std_out *)
            fun propplan_online timeOut st_in = 
		let
		    val problems = Parser.readBatch st_in
		in 
		    (map (process timeOut) problems;())
		end

	    fun getDomain f = let val s = TextIO.openIn f
				  val d = readDomain  (TextIO.getInstream s)
			      in  (TextIO.closeIn s; d) 
			      end

	    fun getProblem f = 
		let val s = TextIO.openIn f
		    val d = readProblem (TextIO.getInstream s)
		in  (TextIO.closeIn s; d) 
                end

            fun pp domain problems =
                map ( process (! time)(*seconds*)) 
                (mkBatch (getDomain domain, map getProblem problems))
	end;
    end;


    
    

